from .loggable import Loggable


class LoggableChild(Loggable):
    parent = None

    def __init__(self, *args, **kwargs):
        parent = kwargs.pop('parent', self.parent)
        if not parent:
            es = "Child object must be passed parent at __init__()"
            raise ValueError(es)
        self.parent = parent
        super().__init__(*args, **kwargs)



from ..utils import str_true, sanitize_dict
from .. import LoggableChild


class LSSTEnvironmentManager(LoggableChild):
    '''Class to hold LSST-specific, but not user-specific, environment.
    The user- specific pieces (from the volume manager and the quota
    manager) will be injected at spawn time.

    The environment should be regarded as immutable after creation.
    '''
    pod_env = {}

    def create_pod_env(self):
        '''Return a dict mapping string to string for injection into the
        pod environment.
        '''
        cfg = self.parent.config
        env = {}
        env['DEBUG'] = str_true(cfg.debug)
        env['MEM_LIMIT'] = cfg.mem_limit
        env['CPU_LIMIT'] = cfg.cpu_limit
        # FIXME
        # Workaround for a bug in our dask templating.
        mem_g = cfg.mem_guarantee
        env['MEM_GUARANTEE'] = self._mem_workaround(mem_g)
        env['CPU_GUARANTEE'] = cfg.cpu_guarantee
        env['LAB_SIZE_RANGE'] = cfg.lab_size_range
        env['CULL_TIMEOUT'] = cfg.cull_timeout
        env['CULL_POLICY'] = cfg.cull_policy
        env['RESTRICT_DASK_NODES'] = str_true(cfg.restrict_dask_nodes)
        env['LAB_NODEJS_MAX_MEM'] = cfg.lab_nodejs_max_mem
        env['NODE_OPTIONS'] = cfg.lab_node_options
        env['EXTERNAL_HUB_URL'] = cfg.external_hub_url
        env['HUB_ROUTE'] = cfg.hub_route
        env['EXTERNAL_HUB_URL'] = cfg.external_hub_url
        env['EXTERNAL_URL'] = cfg.external_hub_url
        env['EXTERNAL_INSTANCE_URL'] = cfg.external_instance_url
        env['FIREFLY_ROUTE'] = cfg.firefly_route
        env['JS9_ROUTE'] = cfg.js9_route
        env['API_ROUTE'] = cfg.api_route
        env['TAP_ROUTE'] = cfg.tap_route
        env['SODA_ROUTE'] = cfg.soda_route
        env['WORKFLOW_ROUTE'] = cfg.workflow_route
        env['EXTERNAL_FIREFLY_URL'] = cfg.external_firefly_url
        env['EXTERNAL_JS9_URL'] = cfg.external_js9_url
        env['EXTERNAL_API_URL'] = cfg.external_api_url
        env['EXTERNAL_TAP_URL'] = cfg.external_tap_url
        env['EXTERNAL_SODA_URL'] = cfg.external_soda_url
        env['EXTERNAL_WORKFLOW_URL'] = cfg.external_workflow_url
        env['AUTO_REPO_URLS'] = cfg.auto_repo_urls
        # Now clean up the env hash by removing any keys with empty values
        cleaned = self._clean_env(env)
        sanitized = self._sanitize(cleaned)
        self.log.debug("create_env yielded:\n.{}".format(sanitized))
        self.pod_env = cleaned

    def _mem_workaround(self, mem):
        '''We need to stop appending "M" to the dask template.  For now
        we return size-in-megabytes-with-no-suffix.
        '''
        if not mem:
            return "1"
        mem_s = str(mem)
        last_c = mem_s[-1].upper()
        # If we get ever-so-precious "*i", deal with it.  "M" also means
        #  2^20 here, not 10^6.  Deal with it.
        if last_c.isdigit():
            return mem_s
        mem_s = mem_s[:-1]
        if last_c == "i":
            mem_s = mem_s[:-1]
            last_c = mem_s[-1]
        mem_i = None
        try:
            mem_i = int(mem_s)
        except ValueError:
            return "1"
        if mem_i < 1:
            mem_i = 1
        if last_c == "K":
            mem_i = int(mem_i / 1024)
            if mem_i < 1:
                mem_i = 1
        elif last_c == "G":
            mem_i = 1024 * mem_i
        elif last_c == "T":
            mem_i = 1024 * 1024 * mem_i
        else:
            # Assume M
            pass
        return str(mem_i)

    def _clean_env(self, env):
        return {str(k): str(v) for k, v in env.items() if (v is not None and
                                                           str(v) != '')}

    def get_env(self):
        '''Return the whole stored environment to caller as a dict.
        '''
        return self.pod_env

    def get_env_key(self, key):
        '''Return value of a specific key in the stored environment to caller.
        '''
        return self.pod_env.get(key)

    def _sanitize(self, incoming):
        sensitive = ['ACCESS_TOKEN', 'GITHUB_ACCESS_TOKEN',
                     'JUPYTERHUB_API_TOKEN', 'JPY_API_TOKEN']
        return sanitize_dict(incoming, sensitive)

    def dump(self):
        '''Return contents as a dict for pretty-printing.
        '''
        ed = {"parent": str(self.parent),
              "pod_env": self._sanitize(self.pod_env)
              }
        return ed



from .authenticator import (LSSTGitHubOAuthenticator,
                            LSSTCILogonOAuthenticator,
                            LSSTJWTAuthenticator)
from .spawner import LSSTSpawner


def configure_auth_and_spawner(config):
    '''Do all the LSST-specific configuration based on the authenticator
    type and environment variables.
    '''
    config.spawner_class = LSSTSpawner
    authclass = None
    authtype = config.authenticator_type
    if authtype == 'jwt':
        authclass = LSSTJWTAuthenticator
        # CILogon is the backing store for JWT at NCSA
        config.allowed_groups = config.cilogon_allowlist
        config.forbidden_groups = config.cilogon_denylist
    elif authtype == 'cilogon':
        authclass = LSSTCILogonOAuthenticator
        authclass.scope = ['openid', 'org.cilogon.userinfo']
        authclass.skin = config.cilogon_skin
        idp = config.cilogon_idp
        if idp:
            authclass.idp = config.cilogon_idp
        config.allowed_groups = config.cilogon_allowlist
        config.forbidden_groups = config.cilogon_denylist
    else:
        authclass = LSSTGitHubOAuthenticator
        config.allowed_groups = config.github_allowlist
        config.forbidden_groups = config.github_denylist
    config.authenticator_class = authclass
    authclass.oauth_callback_url = config.oauth_callback_url
    client_id = config.oauth_client_id
    secret = config.oauth_client_secret
    authclass.client_id = client_id
    authclass.client_secret = secret
    if not client_id:
        config.log.warning("OAuth client ID missing!")
    if not secret:
        config.log.warning("OAuth secret missing!")



import datetime
import functools
import json
import logging
import re
import requests
import semver
import urllib.parse
import urllib.request

from ..utils import make_logger


class ScanRepo(object):
    '''Class to scan repository and create results.

       Based on:
       https://github.com/shangteus/py-dockerhub/blob/master/dockerhub.py'''

    host = 'hub.docker.com'
    path = ''
    owner = ''
    name = ''
    port = None
    data = {}
    debug = False
    json = False
    insecure = False
    sort_field = "name"
    experimentals = 0
    dailies = 3
    weeklies = 2
    releases = 1
    recommended = True
    _results = None
    _results_map = {}
    _name_to_manifest = {}
    _all_tags = []
    logger = None
    cachefile = None

    def __init__(self, host='', path='', owner='', name='',
                 experimentals=0, dailies=3, weeklies=2, releases=1,
                 recommended=True,
                 json=False, port=None,
                 cachefile=None,
                 insecure=False, sort_field="", debug=False):
        if debug:
            self.debug = debug
        self.logger = make_logger()
        if debug:
            self.logger.setLevel(logging.DEBUG)
            self.logger.debug("Debug logging enabled.")
        if host:
            self.host = host
        if path:
            self.path = path
        if owner:
            self.owner = owner
        if name:
            self.name = name
        if experimentals is not None:
            self.experimentals = experimentals
        if dailies is not None:
            self.dailies = dailies
        if weeklies is not None:
            self.weeklies = weeklies
        if releases is not None:
            self.releases = releases
        if json:
            self.json = json
        protocol = "https"
        if insecure:
            self.insecure = insecure
            protocol = "http"
        if sort_field:
            self.sort_field = sort_field
        exthost = self.host
        reghost = exthost
        if reghost == "hub.docker.com":
            reghost = "registry.hub.docker.com"
        if port:
            exthost += ":" + str(port)
            reghost += ":" + str(port)
        if cachefile:
            self.cachefile = cachefile
            self._read_cachefile()
        if not self.path:
            self.path = ("/v2/repositories/" + self.owner + "/" +
                         self.name + "/tags/")
        self.url = protocol + "://" + exthost + self.path
        self.registry_url = (protocol + "://" + reghost + "/v2/" +
                             self.owner + "/" + self.name + "/")
        self.logger.debug("URL: {}".format(self.url))
        self.logger.debug("Registry URL: {}".format(self.registry_url))

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def close(self):
        '''Close the session.
        '''
        if self._session:
            self._session.close()

    def extract_image_info(self):
        '''Build image name list and image description list.
        '''
        cs = []
        if (self.recommended and "recommended" in self.data):
            cs.extend(self.data["recommended"])
        for k in ["experimental", "daily", "weekly", "release"]:
            if k in self.data:
                cs.extend(self.data[k])
        ldescs = []
        for c in cs:
            tag = c["name"]
            ld = c.get("description")
            if not ld:
                ld = self._describe_tag(tag)
            ldescs.append(ld)
        ls = [self.owner + "/" + self.name + ":" + x["name"] for x in cs]
        return ls, ldescs

    def _read_cachefile(self):
        fn = self.cachefile
        try:
            with open(fn) as f:
                data = json.load(f)
        except Exception as exc:
            self.logger.error(
                "Failed to load cachefile '{}'; must rescan".format(fn))
            self.logger.error("Error: {}".format(exc))
            return
        self.logger.debug("Loaded cachefile {}".format(fn))
        nm = self._name_to_manifest
        rm = self._results_map
        for tag in data.keys():
            ihash = data[tag].get("hash")
            ilayers = data[tag].get("layers")
            updated = None
            updatedstr = data[tag].get("updated")
            if updatedstr:
                updated = self._convert_time(updatedstr)
            if ihash and updated:
                if (tag not in nm or (nm[tag]["updated"] < updated)):
                    nm[tag] = {"hash": ihash,
                               "updated": updated}
                    if ilayers:
                        nm[tag]["layers"] = ilayers
                if tag not in rm:
                    rm[tag] = {"last_updated": updatedstr,
                               "name": tag}
                else:
                    l_updated = self._convert_time(rm[tag]["last_updated"])
                    if l_updated < updated:
                        rm[tag] = {"last_updated": updatedstr,
                                   "name": tag}

    def _describe_tag(self, tag):
        # New-style tags have underscores separating components.
        ld = tag  # Default description is just the tag name
        components = None
        if tag.find('_') != -1:
            components = tag.split('_')
            btype = components[0]
            # Handle the r17_0_1 case.
            ctm = re.search(r'\d+$', btype)
            if ctm is not None:
                mj = int(ctm.group())
                components.insert(1, mj)
                btype = btype[0]
            if (tag.startswith("recommended") or tag.startswith("latest")):
                ld = tag[0].upper() + tag[1:]
                restag = self.resolve_tag(tag)
                if restag:
                    ld += " ({})".format(self._describe_tag(restag))
            elif btype == "r":
                rmaj = components[1]
                rmin = components[2]
                rpatch = None
                rrest = None
                if len(components) > 3:
                    rpatch = components[3]
                if len(components) > 4:
                    rrest = "_".join(components[4:])
                ld = "Release %s.%s" % (rmaj, rmin)
                if rpatch:
                    ld = ld + "." + rpatch
                if rrest:
                    ld = ld + "-" + rrest
            elif btype == "w":
                year = components[1]
                week = components[2]
                ld = "Weekly %s_%s" % (year, week)
            elif btype == "d":
                year = components[1]
                month = components[2]
                day = components[3]
                ld = "Daily %s_%s_%s" % (year, month, day)
            elif btype == "exp":
                rest = "_".join(components[1:])
                ld = "Experimental %s" % rest
        else:
            if (tag.startswith("recommended") or tag.startswith("latest")):
                ld = tag[0].upper() + tag[1:]
                restag = self.resolve_tag(tag)
                if restag:
                    ld += " ({})".format(self._describe_tag(restag))
            elif tag[0] == "r":
                rmaj = tag[1:3]
                rmin = tag[3:]
                ld = "Release %s.%s" % (rmaj, rmin)
            elif tag[0] == "w":
                year = tag[1:5]
                week = tag[5:]
                ld = "Weekly %s_%s" % (year, week)
            elif tag[0] == "d":
                year = tag[1:5]
                month = tag[5:7]
                day = tag[7:]
                ld = "Daily %s_%s_%s" % (year, month, day)
            elif tag[0] == "e":
                rest = tag[1:]
                ld = "Experimental %s" % rest
        return ld

    def resolve_tag(self, tag):
        '''Resolve a tag (used for "recommended" or "latest*").
        '''
        mfest = self._name_to_manifest.get(tag)
        if not mfest:
            self.logger.debug("Did not find manifest for '{}'".format(tag))
            return None
        hash = mfest.get("hash")
        self.logger.debug("Tag '{}' hash -> '{}'".format(tag, hash))
        if not hash:
            return None
        for k in self._name_to_manifest:
            if (k.startswith("recommended") or k.startswith("latest")):
                continue
            if self._name_to_manifest[k].get("hash") == hash:
                self.logger.debug("Found matching hash for tag '{}'".format(k))
                return k

    def _data_to_json(self):
        return json.dumps(self.data, sort_keys=True, indent=4,
                          default=self._serialize_datetime)

    def _namemap_to_json(self):
        modmap = {}
        nm = self._name_to_manifest
        rm = self._results_map
        for k in nm:
            dt = nm[k].get("updated")
            dstr = None
            if dt:
                dstr = self._serialize_datetime(dt)
            else:
                dstr = rm[k].get("last_updated")
            ihash = nm[k].get("hash")
            if ihash and dstr:
                modmap[k] = {"updated": dstr,
                             "hash": ihash}
        return json.dumps(modmap, sort_keys=True, indent=4)

    def _serialize_datetime(self, o):
        if isinstance(o, datetime.datetime):
            dstr = o.__str__().replace(' ', 'T')
            if dstr[-1].isdigit():
                dstr += "Z"
            return dstr

    def report(self):
        '''Print the tag data.
        '''
        if self.json:
            print(self._data_to_json())
        else:
            ls, ldescs = self.extract_image_info()
            ldstr = ",".join(ldescs)
            lstr = ",".join(ls)
            print("# Environment variables for Jupyter Lab containers")
            print("LAB_CONTAINER_NAMES=\'%s\'" % lstr)
            print("LAB_CONTAINER_DESCS=\'%s\'" % ldstr)
            print("export LAB_CONTAINER_NAMES LAB_CONTAINER_DESCS")

    def get_data(self):
        '''Return the tag data.
        '''
        return self.data

    def get_all_tags(self):
        '''Return all tags in the repository (sorted by last_updated).
        '''
        return self._all_tags

    def _get_url(self, **kwargs):
        params = None
        resp = None
        url = self.url
        if kwargs:
            params = urllib.parse.urlencode(kwargs)
            url += "?%s" % params
        headers = {"Accept": "application/json"}
        req = urllib.request.Request(url, None, headers)
        resp = urllib.request.urlopen(req)
        page = resp.read()
        return page

    def scan(self):
        '''Perform the repository scan.
        '''
        url = self.url
        self.logger.debug("Beginning repo scan of '{}'.".format(url))
        results = []
        page = 1
        resp_bytes = None
        while True:
            try:
                resp_bytes = self._get_url(page=page)
            except Exception as e:
                message = "Failure retrieving %s: %s" % (url, str(e))
                if resp_bytes:
                    message += " [ data: %s ]" % (
                        str(resp_bytes.decode("utf-8")))
                raise ValueError(message)
            resp_text = resp_bytes.decode("utf-8")
            try:
                j = json.loads(resp_text)
            except ValueError:
                raise ValueError("Could not decode '%s' -> '%s' as JSON" %
                                 (url, str(resp_text)))
            results.extend(j["results"])
            if "next" not in j or not j["next"]:
                break
            page = page + 1
        self._results = results
        self._update_results_map(results)
        self._map_names_to_manifests()
        self._reduce_results()

    def _update_results_map(self, results):
        rm = self._results_map
        for res in results:
            name = res["name"]
            if name not in rm:
                rm[name] = {}
            rm[name].update(res)

    def _map_names_to_manifests(self):
        results = self._results_map
        namemap = self._name_to_manifest
        check_names = []
        for tag in results:
            tstamp = self._convert_time(results[tag]["last_updated"])
            if not namemap.get(tag):
                namemap[tag] = {
                    "layers": None,
                    "updated": tstamp,
                    "hash": None
                }
            if tstamp <= namemap[tag]["updated"] and namemap[tag]["hash"]:
                # We have a manifest
                # Update results map with hash
                results[tag]["hash"] = namemap[tag]["hash"]
                continue
            self.logger.debug("Adding {} to check_names.".format(tag))
            check_names.append(tag)
        if not check_names:
            self.logger.debug("All images have current hash.")
            return
        baseurl = self.registry_url
        url = baseurl + "manifests/recommended"
        i_resp = requests.head(url)
        authtok = None
        sc = i_resp.status_code
        if sc == 401:
            self.logger.debug("Getting token to retrieve layer lists.")
            magicheader = i_resp.headers['Www-Authenticate']
            if magicheader[:7] == "Bearer ":
                hd = {}
                hl = magicheader[7:].split(",")
                for hn in hl:
                    il = hn.split("=")
                    kk = il[0]
                    vv = il[1].replace('"', "")
                    hd[kk] = vv
                if (not hd or "realm" not in hd or "service" not in hd
                        or "scope" not in hd):
                    return None
                endpoint = hd["realm"]
                del hd["realm"]
                tresp = requests.get(endpoint, params=hd, json=True)
                jresp = tresp.json()
                authtok = jresp.get("token")
        elif sc != 200:
            self.logger.warning("GET %s -> %d" % (url, sc))
        # https://docs.docker.com/registry/spec/api/ , "Deleting An Image"
        # Yep, I think that's the only place it tells you that you need
        #  this magic header to get the digest hash.
        headers = {
            "Accept": "application/vnd.docker.distribution.manifest.v2+json"}
        if authtok:
            headers.update({"Authorization": "Bearer {}".format(authtok)})
        for name in check_names:
            resp = requests.head(baseurl + "manifests/{}".format(name),
                                 headers=headers)
            ihash = resp.headers["Docker-Content-Digest"]
            namemap[name]["hash"] = ihash
            results[name]["hash"] = ihash
            dstr = results[name]["last_updated"]
            if dstr:
                dt = self._convert_time(dstr)
                namemap[name]["updated"] = dt
        self._name_to_manifest.update(namemap)
        if self.cachefile:
            self.logger.debug("Writing cache file.")
            try:
                self._writecachefile()
            except Exception as exc:
                self.log.error("Failed to write cache file: {}".format(exc))
            # We're up to date.

    def _writecachefile(self):
        if self.cachefile:
            try:
                with open(self.cachefile, 'w') as f:
                    f.write(self._namemap_to_json())
            except Exception as exc:
                self.logger.error(
                    "Could not write to {}: {}".format(self.cachefile, exc))

    def _reduce_results(self):
        results = self._results
        sort_field = self.sort_field
        # Recommended
        # Release/Weekly/Daily
        # Experimental/Latest/Other
        c_candidates = []
        r_candidates = []
        w_candidates = []
        d_candidates = []
        e_candidates = []
        l_candidates = []
        o_candidates = []
        # This is the order for tags to appear in menu:
        displayorder = []
        if self.recommended:
            displayorder.extend([c_candidates])
        displayorder.extend([e_candidates, d_candidates, w_candidates,
                             r_candidates])
        # This is the order for tags to appear in drop-down:
        imgorder = [l_candidates]
        imgorder.extend(displayorder)
        imgorder.extend(o_candidates)
        reduced_results = {}
        for res in results:
            vname = res["name"]
            reduced_results[vname] = {
                "name": vname,
                "id": res["id"],
                "size": res["full_size"],
                "description": self._describe_tag(vname)
            }
            entry = reduced_results[vname]
            manifest = self._name_to_manifest.get(vname)
            if manifest:
                entry["updated"] = manifest.get("updated")
                entry["hash"] = manifest.get("hash")
            else:
                entry["updated"] = self._convert_time(res["last_updated"])
                entry["hash"] = None
        for res in reduced_results:
            if res.startswith("r") and not res.startswith("recommended"):
                r_candidates.append(reduced_results[res])
            elif res.startswith("w"):
                w_candidates.append(reduced_results[res])
            elif res.startswith("d"):
                d_candidates.append(reduced_results[res])
            elif res.startswith("exp"):
                e_candidates.append(reduced_results[res])
            elif res.startswith("latest"):
                l_candidates.append(reduced_results[res])
            elif res.startswith("recommended"):
                c_candidates.append(reduced_results[res])
            else:
                o_candidates.append(res)
        for clist in imgorder:
            if sort_field != 'name':
                clist.sort(key=lambda x: x[sort_field], reverse=True)
            else:
                clist = self._sort_images_by_name(clist)
        r = {}
        # Index corresponds to order in displayorder
        idxbase = 0
        imap = {}
        if self.recommended:
            imap.update({"recommended": {"index": idxbase,
                                         "count": 1}
                         })
            idxbase = 1
        imap.update({"experimental": {"index": idxbase,
                                      "count": self.experimentals},
                     "daily": {"index": idxbase + 1,
                               "count": self.dailies},
                     "weekly": {"index": idxbase + 2,
                                "count": self.weeklies},
                     "release": {"index": idxbase + 3,
                                 "count": self.releases}
                     })
        for ikey in list(imap.keys()):
            idx = imap[ikey]["index"]
            ict = imap[ikey]["count"]
            if ict:
                r[ikey] = displayorder[idx][:ict]
        all_tags = self._sort_tags_by_date()
        self._all_tags = all_tags
        self.data = r

    def _sort_tags_by_date(self):
        items = [x[1] for x in self._results_map.items()]
        dec = [(x['last_updated'], x['name']) for x in items]
        dec.sort(reverse=True)
        tags = [x[1] for x in dec]
        return tags

    def _sort_images_by_name(self, clist):
        # We have a flag day where we start putting underscores into
        #  image tags.  Those always go at the top.
        # We begin by splitting the list of candidate images into new
        #  and old style images.
        oldstyle = []
        newstyle = []
        for cimg in clist:
            name = cimg["name"]
            if name.find("_") == -1:
                oldstyle.append(cimg)
            else:
                # "latest_X" is not a semantic version tag.
                if name.startswith("latest_"):
                    oldstyle.append(cimg)
                else:
                    newstyle.append(cimg)
        # Old-style sort is simple string comparison.
        oldstyle.sort(key=lambda x: x["name"], reverse=True)
        # New style, we refer to semver module for comparison.
        #  (also works fine for date sorts)
        seml = []
        for cimg in newstyle:
            name = cimg["name"]
            components = name.split("_")
            # Get this.  It's not represented as r_17, no, it's r17.
            # So if we find that the end of the first group is digits,
            #  we split those off with a regular expression, and insert
            #  them into the list where the major number should be.
            ctype = components[0]
            ctm = re.search(r'\d+$', ctype)
            if ctm is not None:
                mj = int(ctm.group())
                components.insert(1, mj)
            # First character is image type, not semantically significant
            #  for versioning.
            if components[0] == "exp":
                _ = components.pop(0)
            major = 0
            if len(components) > 1:
                major = int(components[1])
            minor = 0
            if len(components) > 2:
                minor = int(components[2])
            patch = 0
            prerelease = None
            if len(components) > 3:
                try:
                    patch = int(components[3])
                except ValueError:
                    # Not an integer, so this is probably an experimental/
                    #  not-for-release version, so leave it at patch level 0
                    #  and treat the string as a prerelease version
                    prerelease = components[3]
            if len(components) > 4:
                prerelease = components[4]
            build = None
            if len(components) > 5:
                build = '_'.join(components[5:])
            cimg["semver"] = semver.format_version(
                major, minor, patch, prerelease, build)
            seml.append(cimg["semver"])
        seml.sort(key=functools.cmp_to_key(semver.compare), reverse=True)
        sorted_newstyle = []
        for skey in seml:
            for ni in newstyle:
                if ni["semver"] == skey:
                    sorted_newstyle.append(ni)
                    break
        # Return all new style names first.
        return sorted_newstyle.extend(oldstyle)

    def _sort_releases_by_name(self, r_candidates):
        # rXYZrc2 should *precede* rXYZ
        # We're going to decorate short (that is, no rc tag) release names
        #  with "zzz", re-sort, and then undecorate.
        nm = {}
        for c in r_candidates:
            tag = c["name"]
            if len(tag) == 4:
                xtag = tag+"zzz"
                nm[xtag] = tag
                c["name"] = xtag
        r_candidates.sort(key=lambda x: x["name"], reverse=True)
        for c in r_candidates:
            xtag = c["name"]
            c["name"] = nm[xtag]
        return r_candidates

    def _convert_time(self, ts):
        f = '%Y-%m-%dT%H:%M:%S.%f%Z'
        if ts[-1] == "Z":
            ts = ts[:-1] + "UTC"
        if ts[-1].isdigit():
            # Naive time
            f = '%Y-%m-%dT%H:%M:%S.%f'
        return datetime.datetime.strptime(ts, f)



import datetime
import threading
import time
from .scanrepo import ScanRepo
from ..singleton import Singleton


class SingletonScanner(ScanRepo, metaclass=Singleton):
    '''Singleton Object to hold rate-limited scanner.
    '''
    min_refresh_time = 60
    max_cache_age = 600
    last_updated = datetime.datetime(1970, 1, 1)  # The Epoch
    scanning = False
    lock = threading.RLock()

    def __init__(self, **kwargs):
        min_refresh_time = kwargs.get('min_refresh_time')
        if min_refresh_time is None:
            min_refresh_time = 60
        max_cache_age = kwargs.get('max_cache_age')
        if max_cache_age is None:
            max_cache_age = 600
        # Now remove them from kwargs before superclass init
        for karg in ['min_refresh_time', 'max_cache_age']:
            if karg in kwargs:
                kwargs.pop(karg)
        super().__init__(**kwargs)
        self.min_refresh_time = min_refresh_time
        self.max_cache_age = max_cache_age
        if max_cache_age < min_refresh_time:
            max_cache_age = 2 * min_refresh_time
            self.max_cache_age = max_cache_age
            self.logger.error("Nonsensical cache age/refresh time ratio.")
            self.logger.warning("Setting max_age to %ds." % max_cache_age)
        thd = threading.Thread(target=self.scan)
        self.logger.info("Starting background scan.")
        thd.start()

    def scan(self):
        '''Execute repo scan.
        '''
        _timeout = 300
        _initialdelay = 0.2
        _maxdelay = 15
        _sofar = 0
        delay = _initialdelay
        if self.scanning:
            while not self._results:
                # If there have been no results, wait up to _timeout seconds
                #  for them.
                self.logger.debug(
                    "Scan in progress; waiting {}s for results.".format(delay))
                time.sleep(delay)
                _sofar += delay
                if _sofar >= _timeout:
                    raise ValueError("Scan in progress never finished.")
                delay *= 2
                if delay > _maxdelay:
                    delay = _maxdelay
                if _sofar + delay > _timeout:
                    delay = _timeout - _sofar
            return  # Use cached (possibly stale) version if results exist.
        self.scanning = True
        now = datetime.datetime.utcnow()
        mt = self.min_refresh_time
        min_delay = datetime.timedelta(seconds=mt)
        if (now - self.last_updated < min_delay):
            self.logger.warning("%ds not elapsed; not rescanning." % mt)
            self.scanning = False
            return
        self.logger.info("Rescanning.")
        with self.lock:
            super().scan()
            self.last_updated = now
            self.scanning = False

    def _scan_if_needed(self):
        now = datetime.datetime.utcnow()
        max_age = datetime.timedelta(seconds=self.max_cache_age)
        last_updated = self.last_updated
        if ((now - last_updated) > max_age):
            self.logger.info("Scan data has expired.")
            if self.scanning:
                self.logger.info("Waiting for scan results.")
                count = 0
                while True:
                    time.sleep(0.2)
                    count = count + 1
                    if not (count % 50):
                        self.logger.debug(
                            "Waiting for data: {}s".format(count / 5))
                    if self.data:
                        break
            else:
                self.logger.info("Beginning new scan.")
                self.scan()
            self.logger.debug("Exiting _scan_if_needed() wait loop.")

    def get_data(self):
        '''Return repo data.
        '''
        self._scan_if_needed()
        return self.data

    def get_all_tags(self):
        '''Return all tags in repo.
        '''
        self._scan_if_needed()
        return self._all_tags

    def get_all_scan_results(self):
        '''Return results from repository scan as dict.
        '''
        self._scan_if_needed()
        return self._results_map

    def extract_image_info(self):
        '''Get info for all images.
        '''
        self._scan_if_needed()
        return super().extract_image_info()

    def report(self):
        '''Report results of scan.
        '''
        self._scan_if_needed()
        return super().report()



from .. import LoggableChild
from .. import Singleton
from kubespawner.clients import shared_client
from kubernetes.config import load_incluster_config, load_kube_config
from kubernetes.config.config_exception import ConfigException
from argo.workflows.client import V1alpha1Api
import argo.workflows.config


class LSSTAPIManager(LoggableChild, metaclass=Singleton):
    wf_api = None
    rbac_api = None
    api = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        try:
            load_incluster_config()
            argo.workflows.config.load_incluster_config()
        except ConfigException:
            self.log.warning("In-cluster config failed! Falling back.")
            try:
                load_kube_config()
                argo.workflows.config.load_kube_config()
            except ValueError as exc:
                self.log.error("Still errored: {}".format(exc))
        wf_api = kwargs.pop('wf_api', self.wf_api)
        if not wf_api:
            wf_api = V1alpha1Api()
        self.wf_api = wf_api
        rbac_api = kwargs.pop('rbac_api', self.rbac_api)
        if not rbac_api:
            rbac_api = shared_client('RbacAuthorizationV1Api')
        self.rbac_api = rbac_api
        api = kwargs.pop('api', self.api)
        if not api:
            api = shared_client('CoreV1Api')
        self.api = api

    def dump(self):
        '''Return contents dict for aggregation and pretty-printing.
        '''
        ad = {"parent": str(self.parent),
              "api": str(self.api),
              "rbac_api": str(self.rbac_api),
              "wf_api": str(self.wf_api)
              }
        return ad



'''Class to manage an LSST-specific options form.
'''
import datetime
import jinja2
import json
from collections import OrderedDict
from time import sleep
from .. import SingletonScanner
from .. import LoggableChild


class LSSTOptionsFormManager(LoggableChild):
    '''Class to create and read a spawner form.
    '''
    sizemap = {}
    _scanner = None
    options_form_data = None

    def get_options_form(self):
        '''Create an LSST Options Form from parent's config object.
        '''
        # Make options form by scanning container repository, then cache it.
        # For a single spawning session, you always get the same form.
        #
        # If that's not OK (long-lived tokens, for example) then in
        #  your authenticator's refresh_user(), clear options_form_data.
        self.log.debug("Creating options form.")
        if self.options_form_data:
            self.log.debug("Returning cached options form.")
            return self.options_form_data
        cfg = self.parent.config
        scanner = SingletonScanner(host=cfg.lab_repo_host,
                                   owner=cfg.lab_repo_owner,
                                   name=cfg.lab_repo_name,
                                   experimentals=cfg.prepuller_experimentals,
                                   dailies=cfg.prepuller_dailies,
                                   weeklies=cfg.prepuller_weeklies,
                                   releases=cfg.prepuller_releases,
                                   cachefile=cfg.prepuller_cachefile,
                                   debug=cfg.debug)
        self._scanner = scanner
        self._sync_scan()
        lnames, ldescs = scanner.extract_image_info()
        desclist = []
        # Setting this up to pass into the Jinja template more easily
        for idx, img in enumerate(lnames):
            desclist.append({"name": img,
                             "desc": ldescs[idx]})
        colon = lnames[0].find(':')
        custtag = lnames[0][:colon] + ":__custom"
        all_tags = scanner.get_all_tags()
        now = datetime.datetime.now()
        nowstr = now.ctime()
        if not now.tzinfo:
            # If we don't have tzinfo, assume it's in UTC
            nowstr += " UTC"
        self._make_sizemap()
        # in order to get the right default size index, we need to poke the
        #  quota manager, because different users may get different sizes
        #  by default
        cfg = self.parent.config
        qm = self.parent.quota_mgr
        qm.define_resource_quota_spec()
        defaultsize = qm.custom_resources.get('size_index') or cfg.size_index
        template_file = self.parent.config.form_template
        template_loader = jinja2.FileSystemLoader(searchpath='/')
        template_environment = jinja2.Environment(loader=template_loader)
        template = template_environment.get_template(template_file)
        optform = template.render(
            defaultsize=defaultsize,
            desclist=desclist,
            all_tags=all_tags,
            custtag=custtag,
            sizelist=list(self.sizemap.values()),
            nowstr=nowstr)
        self.options_form_data = optform
        return optform

    def resolve_tag(self, tag):
        '''Delegate to scanner to resolve convenience tags.
        '''
        return self._scanner.resolve_tag(tag)

    def _sync_scan(self):
        scanner = self._scanner
        cfg = self.parent.config
        delay_interval = cfg.initial_scan_interval
        max_delay_interval = cfg.max_scan_interval
        max_delay = cfg.max_scan_delay
        delay = 0
        epoch = datetime.datetime(1970, 1, 1)
        while scanner.last_updated == epoch:
            self.log.info(("Scan results not available yet; sleeping " +
                           "{:02.1f}s ({:02.1f}s " +
                           "so far).").format(delay_interval, delay))
            sleep(delay_interval)
            delay = delay + delay_interval
            delay_interval *= 2
            if delay_interval > max_delay_interval:
                delay_interval = max_delay_interval
            if delay >= max_delay:
                errstr = ("Scan results did not become available in " +
                          "{}s.".format(max_delay))
                raise RuntimeError(errstr)

    def _make_sizemap(self):
        sizemap = OrderedDict()
        # For supported Python versions, dict is ordered anyway...
        sizes = self.parent.config.form_sizelist
        tiny_cpu = self.parent.config.tiny_cpu
        mem_per_cpu = self.parent.config.mb_per_cpu
        # Each size doubles the previous one.
        cpu = tiny_cpu
        idx = 0
        for sz in sizes:
            mem = mem_per_cpu * cpu
            sizemap[sz] = {"cpu": cpu,
                           "mem": mem,
                           "name": sz,
                           "index": idx}
            desc = sz.title() + " (%.2f CPU, %dM RAM)" % (cpu, mem)
            sizemap[sz]["desc"] = desc
            cpu = cpu * 2
            idx = idx + 1
        # Clean up if list of sizes changed.
        self.sizemap = sizemap

    def options_from_form(self, formdata=None):
        '''Get user selections.
        '''
        options = None
        if formdata:
            self.log.debug("Form data: %s", json.dumps(formdata,
                                                       sort_keys=True,
                                                       indent=4))
            options = {}
            if ('kernel_image' in formdata and formdata['kernel_image']):
                options['kernel_image'] = formdata['kernel_image'][0]
            if ('size' in formdata and formdata['size']):
                options['size'] = formdata['size'][0]
            if ('image_tag' in formdata and formdata['image_tag']):
                options['image_tag'] = formdata['image_tag'][0]
            if ('clear_dotlocal' in formdata and formdata['clear_dotlocal']):
                options['clear_dotlocal'] = True
        return options

    def dump(self):
        '''Return contents dict for pretty-printing.
        '''
        sd = {"parent": str(self.parent),
              "sizemap": self.sizemap,
              "scanner": str(self._scanner)}
        return sd



from ..utils import make_logger


class Loggable(object):
    log = None

    def __init__(self, *args, **kwargs):
        self.log = make_logger()
        self.log.debug("Creating {}".format(self.__class__.__name__))



from .. import Loggable
from .apimanager import LSSTAPIManager
from .authmanager import LSSTAuthManager
from .envmanager import LSSTEnvironmentManager
from .namespacemanager import LSSTNamespaceManager
from .optionsformmanager import LSSTOptionsFormManager
from .quotamanager import LSSTQuotaManager
from .volumemanager import LSSTVolumeManager
from .workflowmanager import LSSTWorkflowManager


class LSSTMiddleManager(Loggable):
    '''The LSSTMiddleManager is a class that holds references to various
    LSST-specific management objects and delegates requests to them.
    The idea is that an LSST Spawner, or an LSST Workflow Manager,
    could instantiate a single LSSTMiddleManager, which would then be
    empowered to perform all LSST-specific operations, reducing
    configuration complexity.
    '''
    parent = None
    config = None
    authenticator = None
    spawner = None
    user = None
    api = None
    rbac_api = None
    wf_api = None
    api_mgr = None
    auth_mgr = None
    namespace_mgr = None
    optionsform_mgr = None
    quota_mgr = None
    volume_mgr = None
    workflow_mgr = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.log.info(
            "Parent of LSST Middle Manager is '{}'".format(self.parent))
        self.config = kwargs.pop('config')
        parent = kwargs.pop('parent', self.parent)
        self.parent = parent
        authenticator = kwargs.pop('authenticator', self.authenticator)
        self.authenticator = authenticator
        self.authenticator = self.parent
        self.api_mgr = LSSTAPIManager(parent=self)
        self.auth_mgr = LSSTAuthManager(parent=self)
        self.env_mgr = LSSTEnvironmentManager(parent=self)
        self.namespace_mgr = LSSTNamespaceManager(parent=self)
        self.optionsform_mgr = LSSTOptionsFormManager(parent=self)
        self.quota_mgr = LSSTQuotaManager(parent=self)
        self.volume_mgr = LSSTVolumeManager(parent=self)
        self.workflow_mgr = LSSTWorkflowManager(parent=self)
        self.api = self.api_mgr.api
        self.rbac_api = self.api_mgr.rbac_api
        self.wf_api = self.api_mgr.wf_api

    def ensure_resources(self):
        '''Delegate to namespace manager.
        '''
        self.namespace_mgr.ensure_namespace()

    def dump(self):
        '''Return contents dict to pretty-print.
        '''
        md = {"parent": str(self.parent),
              "authenticator": str(self.authenticator),
              "spawner": str(self.spawner),
              "user": str(self.user),
              "api": str(self.api),
              "rbac_api": str(self.rbac_api),
              "wf_api": str(self.wf_api),
              "config": self.config.dump(),
              "api_mgr": self.api_mgr.dump(),
              "auth_mgr": self.auth_mgr.dump(),
              "env_mgr": self.env_mgr.dump(),
              "optionsform_mgr": self.optionsform_mgr.dump(),
              "quota_mgr": self.quota_mgr.dump(),
              "volume_mgr": self.volume_mgr.dump(),
              "workflow_mgr": self.workflow_mgr.dump()
              }
        return md



import base64
import json
import os
import yaml
from kubernetes import client
from .. import LoggableChild


class LSSTVolumeManager(LoggableChild):
    '''Class to provide support for document-driven Volume assignment.
    '''
    volume_list = []
    k8s_volumes = []
    k8s_vol_mts = []

    def make_volumes_from_config(self):
        '''Create volume definition representation from document.
        Override this in a subclass if you like.
        '''
        vollist = []
        config = []
        config_file = self.parent.config.volume_definition_file
        if not os.path.exists(config_file):
            return vollist
        with open(config_file, "r") as fp:
            config = json.load(fp)
        for mtpt in config:
            mountpoint = mtpt["mountpoint"]  # Fatal error if it doesn't exist
            if mtpt.get("disabled"):
                continue
            if mountpoint[0] != "/":
                mountpoint = "/" + mountpoint
            host = (mtpt.get("fileserver-host") or
                    self.parent.config.fileserver_host)
            export = mtpt.get("fileserver-export") or (
                "/exports" + mountpoint)
            mode = (mtpt.get("mode") or "ro").lower()
            k8s_vol = mtpt.get("kubernetes-volume")
            if k8s_vol:
                raise ValueError("Shadow PVs and matching PVCs " +
                                 "are no longer supported!")
            hostpath = mtpt.get("hostpath")
            vollist.append({
                "mountpoint": mountpoint,
                "hostpath": hostpath,
                "host": host,
                "export": export,
                "mode": mode,
            })
        self.volume_list = vollist
        self.log.debug("Volumes: {}".format(vollist))
        self._define_k8s_object_representations()

    def _define_k8s_object_representations(self):
        self.k8s_volumes = []
        self.k8s_vol_mts = []
        for vol in self.volume_list:
            k8svol = None
            k8smt = None
            if vol.get("hostpath"):
                k8svol = self._define_k8s_hostpath_vol(vol)
            else:
                k8svol = self._define_k8s_nfs_vol(vol)
            k8smt = self._define_k8s_mtpt(vol)
            if k8svol and k8smt:
                self.k8s_volumes.append(k8svol)
                self.k8s_vol_mts.append(k8smt)

    def _define_k8s_hostpath_vol(self, vol):
        return client.V1Volume(
            name=self._get_volume_name_for_mountpoint(vol["mountpoint"]),
            host_path=client.V1HostPath(
                path=vol["mountpoint"]
            )
        )

    def _define_k8s_nfs_vol(self, vol):
        knf = client.V1NFSVolumeSource(
            path=vol["export"],
            server=vol["host"]
        )
        if vol["mode"] == "ro":
            knf.read_only = True
        return client.V1Volume(
            name=self._get_volume_name_for_mountpoint(vol["mountpoint"]),
            nfs=knf
        )

    def _define_k8s_mtpt(self, vol):
        mt = client.V1VolumeMount(
            mount_path=vol["mountpoint"],
            name=self._get_volume_name_for_mountpoint(vol["mountpoint"]),
        )
        if vol["mode"] == "ro":
            mt.read_only = True
        return mt

    def _get_volume_name_for_mountpoint(self, mountpoint):
        return mountpoint[1:].replace('/', '-')

    def _get_volume_yaml_str(self, left_pad=0):
        vols = self.k8s_volumes
        if not vols:
            self.log.warning("No volumes defined.")
            return ''
        vl = []
        for vol in vols:
            nm = vol.name
            hp = vol.host_path
            nf = vol.nfs
            vo = {"name": nm}
            if hp:
                vo["hostPath"] = {"path": hp.path}
            elif nf:
                am = "ReadWriteMany"
                if nf.read_only:
                    am = "ReadOnlyMany"
                vo["nfs"] = {"server": nf.server,
                             "path": nf.path,
                             "accessMode": am}
            vl.append(vo)
        vs = {"volumes": vl}
        ystr = yaml.dump(vs)
        return self._left_pad(ystr, left_pad)

    def _get_volume_mount_yaml_str(self, left_pad=0):
        vms = self.k8s_vol_mts
        if not vms:
            self.log.warning("No volume mounts defined.")
            return ''
        vl = []
        for vm in vms:
            vo = {}
            vo["name"] = vm.name
            vo["mountPath"] = vm.mount_path
            if vm.read_only:
                vo["readOnly"] = True
            vl.append(vo)
        vs = {"volumeMounts": vl}
        ystr = yaml.dump(vs)
        return self._left_pad(ystr, left_pad)

    def _left_pad(self, line_str, left_pad=0):
        pad = ' ' * left_pad
        ylines = line_str.split("\n")
        padlines = [pad + l for l in ylines]
        return "\n".join(padlines)

    def get_dask_volume_b64(self):
        '''Return the base-64 encoding of the K8s statements to create
        the pod's mountpoints.  Probably better handled as a ConfigMap.
        '''
        vmt_yaml_str = self._get_volume_mount_yaml_str(left_pad=4)
        vol_yaml_str = self._get_volume_yaml_str(left_pad=2)
        ystr = "{}\n{}".format(vmt_yaml_str, vol_yaml_str)
        benc = base64.b64encode(ystr.encode('utf-8')).decode('utf-8')
        return benc

    def dump(self):
        '''Return contents dict for aggregation and pretty-printing.
        '''
        vd = {"parent": str(self.parent),
              "volume_list": self.volume_list,
              "k8s_volumes": [str(x) for x in self.k8s_volumes],
              "k8s_vol_mts": [str(x) for x in self.k8s_vol_mts]}
        return vd



'''
Shared utility functions.
'''

import hashlib
import logging
import os

from collections import defaultdict


def rreplace(s, old, new, occurrence):
    '''Convenience function from:
    https://stackoverflow.com/questions/2556108/\
    rreplace-how-to-replace-the-last-occurrence-of-an-expression-in-a-string
    '''
    li = s.rsplit(old, occurrence)
    return new.join(li)


def sanitize_dict(input_dict, sensitive_fields):
    retval = {}
    if not input_dict:
        return retval
    retval.update(input_dict)
    for field in sensitive_fields:
        if retval.get(field):
            retval[field] = "[redacted]"
    return retval


def get_execution_namespace():
    '''Return Kubernetes namespace of this container.
    '''
    ns_path = '/var/run/secrets/kubernetes.io/serviceaccount/namespace'
    if os.path.exists(ns_path):
        with open(ns_path) as f:
            return f.read().strip()
    return None


def make_logger(name=__name__):
    '''Create a logger with a specific output format.
    '''
    logger = logging.getLogger(name)
    fstr = '[%(levelname).1s %(asctime)s.%(msecs).03d'
    fstr += ' %(module)s:%(funcName)s:%(lineno)d] %(message)s'
    dstr = '%Y-%m-%d %H:%M:%S'
    ch = logging.StreamHandler()
    fmt = logging.Formatter(fmt=fstr, datefmt=dstr)
    ch.setFormatter(fmt)
    # Remove default handlers, if any
    logger.handlers = []
    logger.addHandler(ch)
    return logger


def str_bool(s):
    '''Make a sane guess for whether a value represents true or false.
    Intended for strings, mostly in the context of environment variables,
    but if you pass it something that's not a string that is falsy, like
    an empty list, it will cheerfully return False.
    '''
    if not s:
        return False
    if type(s) != str:
        # It's not a string and it's not falsy, soooo....
        return True
    s = s.lower()
    if s in ['false', '0', 'no', 'n']:
        return False
    return True


def str_true(v):
    '''The string representation of a true value will be 'TRUE'.  False will
    be the empty string.
    '''
    if v:
        return 'TRUE'
    else:
        return ''


def listify(item, delimiter=','):
    '''Used for taking character (usually comma)-separated string lists
    and returning an actual list, or the empty list.
    Useful for environment parsing.

    Sure, you could pass it integer zero and get [] back.  Don't.
    '''
    if not item:
        return []
    if type(item) is str:
        item = item.split(delimiter)
    if type(item) is not list:
        raise TypeError("'listify' must take None, str, or list!")
    return item


def floatify(item, default=0.0):
    '''Another environment-parser: the empty string should be treated as
    None, and return the default, rather than the empty string (which
    does not become an integer).  Default can be either a float or string
    that float() works on.  Note that numeric zero (or string '0') returns
    0.0, not the default.  This is intentional.
    '''
    if item is None:
        return default
    if item == '':
        return default
    return float(item)


def intify(item, default=0):
    '''floatify, but for ints.
    '''
    return int(floatify(item, default))


def list_duplicates(seq):
    '''List duplicate items from a sequence.
    '''
    # https://stackoverflow.com/questions/5419204
    tally = defaultdict(list)
    for i, item in enumerate(seq):
        tally[item].append(i)
    return ((key, locs) for key, locs in tally.items()
            if len(locs) > 1)


def list_digest(inp_list):
    if type(inp_list) is not list:
        raise TypeError("list_digest only works on lists!")
    if not inp_list:
        raise ValueError("input must be a non-empty list!")
    # If we can rely on python >= 3.8, shlex.join is better
    return hashlib.sha256(' '.join(inp_list).encode('utf-8')).hexdigest()



'''LSST Login Handler to use JWT token present in request headers.
'''
import datetime
from tornado import gen, web
from jupyterhub.utils import url_path_join
from jwtauthenticator.jwtauthenticator import JSONWebTokenLoginHandler
from ..lsstmgr import check_membership


class LSSTJWTLoginHandler(JSONWebTokenLoginHandler):

    @gen.coroutine
    def get(self):
        '''Authenticate on get() via reading the token from HTTP headers.
        '''
        # This is taken from https://github.com/mogthesprog/jwtauthenticator
        #  but with our additional claim information checked and stuffed
        #  into auth_state, and allow/deny lists checked.
        #
        # We can't just use the superclass because it issues the redirect,
        #  and we can't do that until we do our additional checking.
        claims, token = yield self._check_auth_header()
        username_claim_field = self.authenticator.username_claim_field
        username = self.retrieve_username(claims, username_claim_field)
        user = self.user_from_username(username)
        # Here is where we deviate from the vanilla JWT authenticator.
        # We simply store all the JWT claims in auth_state, although we also
        #  choose our field names to make the spawner reusable from the
        #  OAuthenticator implementation.
        auth_state = yield self.refresh_user(user)
        if not self._jwt_validate_user(auth_state):
            # We're either in a forbidden group, or not in any allowed group
            self.log.error("User did not validate from claims groups.")
            raise web.HTTPError(403)
        _ = yield user.save_auth_state(auth_state)
        self.set_login_cookie(user)
        _url = url_path_join(self.hub.server.base_url, 'home')
        next_url = self.get_argument('next', default=False)
        if next_url:
            _url = next_url
        self.redirect(_url)

    @gen.coroutine
    def refresh_user(self, user, handler=None):
        '''Validate the token and force re-auth if the claims are not
        (presumably no longer) valid.
        '''
        self.log.debug("Refreshing user data.")
        try:
            claims, token = yield self._check_auth_header()
        except web.HTTPError:
            # Force re-login
            return False
        username_claim_field = self.authenticator.username_claim_field
        username = self.retrieve_username(claims, username_claim_field)
        auth_state = {"id": username,
                      "access_token": token,
                      "claims": claims}
        return auth_state

    @gen.coroutine
    def _check_auth_header(self):
        # Either returns (valid) claims and token,
        #  or throws a web error of some type.
        self.log.debug("Checking authentication header.")
        cfg = self.authenticator.lsst_mgr.config
        header_name = self.authenticator.header_name
        param_name = self.authenticator.param_name
        header_is_authorization = self.authenticator.header_is_authorization
        auth_header_content = self.request.headers.get(header_name, "")
        auth_cookie_content = self.get_cookie("XSRF-TOKEN", "")
        signing_certificate = cfg.jwt_signing_certificate
        secret = self.authenticator.secret
        audience = cfg.audience
        tokenParam = self.get_argument(param_name, default=False)
        if auth_header_content and tokenParam:
            self.log.error("Authentication: both an authentication header " +
                           "and tokenParam")
            raise web.HTTPError(400)
        elif auth_header_content:
            if header_is_authorization:
                # We should not see "token" as first word in the
                #  authorization header.  If we do it could mean someone
                #  coming in with a stale API token
                if auth_header_content.split()[0].lower() != "bearer":
                    self.log.error("Authorization header is not 'bearer'.")
                    raise web.HTTPError(403)
                token = auth_header_content.split()[1]
            else:
                token = auth_header_content
        elif auth_cookie_content:
            token = auth_cookie_content
        elif tokenParam:
            token = tokenParam
        else:
            self.log.error("Could not determine authentication token.")
            raise web.HTTPError(401)
        claims = ""
        if secret:
            claims = self.verify_jwt_using_secret(token, secret, audience)
        elif signing_certificate:
            claims = self.verify_jwt_with_claims(token, signing_certificate,
                                                 audience)
        else:
            self.log.error("Could not verify JWT.")
            raise web.HTTPError(401)
        # Check expiration
        expiry = int(claims['exp'])
        now = int(datetime.datetime.utcnow().timestamp())
        if now > expiry:
            self.log.error("JWT has expired!")
            raise web.HTTPError(401)
        return claims, token

    def _jwt_validate_user(self, auth_state):
        cfg = self.authenticator.lsst_mgr.config
        claims = auth_state['claims']
        uid, groupmap = self.authenticator.resolve_cilogon(claims)
        groups = list(groupmap.keys())
        valid = check_membership(
            groups, cfg.allowed_groups, cfg.forbidden_groups, log=self.log)
        if not valid:
            self.log.warning("Group membership check failed.")
            return False
        auth_state['group_map'] = groupmap
        auth_state['uid'] = uid
        return True
